/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "fh2805config.h"

#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>

#include <portLib.h>

#include "fh2805/fh2805.h"
#include "fh2805Cmds.h"

static int getIntUnsupported(fh2805dev_t *);
static int getVelocity(fh2805dev_t *);
static int getHomingStatus(fh2805dev_t *);
static int getMaxSpeed(fh2805dev_t *);
static int getAcceleration(fh2805dev_t *);
static int getPropTerm(fh2805dev_t *);
static int getIntTerm(fh2805dev_t *);
static int getMode(fh2805dev_t *);
static int getProgramSequence(fh2805dev_t *);
static int getPos(fh2805dev_t *);
static int getSpeed(fh2805dev_t *);
static int getTemp(fh2805dev_t *);
static int getVersion(fh2805dev_t *);
static int getStatus(fh2805dev_t *);
static int getActualStatus(fh2805dev_t *);

#ifndef NUL
# define NUL '\0'
#endif

/*
 * Table of commands for the motor controller:
 * - symbolic name
 * - string to send to the controller (with format specifier for parameters)
 * - expected asynchronous message, or 'NUL' if none
 * - function used to decode the reply
 */
static FH2805_CMD_TAB cmdTab[] = {
/* Basic settings */
	{ DISABLE_DRIVE, "DI", NUL, NULL },
	{ ENABLE_DRIVE, "EN", NUL, NULL },
	{ SET_BAUD, "BAUD%u", NUL, NULL },
	{ SET_NODE_ADDRESS, "NODEADR%u", NUL, NULL },
	{ GET_NODE_ADDRESS, "GNODEADR", NUL, getIntUnsupported },
	{ SET_ENCODER_RESOLUTION, "ENCRES%u", NUL, NULL },
	{ GET_ENCODER_RESOLUTION, "GENCRES", NUL, getIntUnsupported },
	{ SET_ASYNC_ANSWER, "ANSW%u", NUL, NULL }, 
	{ SET_POSITION_RANGE_LIMITS, "LL%d", NUL, NULL },
	{ ACTIVATE_POSITION_LIMITS, "APL%u", NUL, NULL },
	{ GET_POSITIVE_LIMIT, "GPL", NUL, getIntUnsupported },
	{ GET_NEGATIVE_LIMIT, "GNL", NUL, getIntUnsupported }, 
	{ RESET_NODE, "RN", NUL, NULL }, 
	{ SAVE_TO_EEPROM, "EEPSAV", NUL, NULL },
	{ RESET_FACTORY_CONFIG, "FCONFIG", NUL, NULL },
	                                        
/* Motion control */				
	{ ENABLE_POSITIONNING_MODE, "M", NUL, NULL },
	{ SET_ABSOLUTE_POSITION_REFERENCE, "LA%d", NUL, NULL },
	{ SET_RELATIVE_POSITION_REFERENCE, "LR%d", NUL, NULL },
	{ ENABLE_VELOCITY_MODE, "V%d", NUL, NULL },
	{ GET_VELOCITY, "GV", NUL, getVelocity }, 
	{ NOTIFY_POSITION, "NP%d", 'p', NULL },
	{ NOTIFY_POSITION_GOAL, "NP", 'p', NULL },
	{ NOTIFY_VELOCITY, "NV%d", 'v', NULL }, 
	{ NOTIFY_POSITION_OFF, "NPOFF", NUL, NULL },
	{ NOTIFY_VELOCITY_OFF, "NVOFF", NUL, NULL },
	                                        
/* Velocity control */				
	{ SET_MINIMUM_VELOCITY, "MV%u", NUL, NULL }, 
	{ GET_MINIMUM_VELOCITY, "GMV", NUL, getIntUnsupported }, 
	{ SET_MINIMUM_ANALOG_VOLTAGE, "MAV%u", NUL, NULL },
	{ GET_MINIMUM_ANALOG_VOLDATE, "GMAV", NUL, getIntUnsupported },
	{ SET_ANALOG_DIRECTION_LEFT, "ADL", NUL, NULL },
	{ SET_ANALOG_DIRECTION_RIGHT, "ADR", NUL, NULL },
	{ SET_VELOCITY_SOURCE, "SOR%d", NUL, NULL },
	                                        
/* Homing points and limit switches */		
	{ SET_POSITION_ZERO, "HO", NUL, NULL },
	{ SET_POSITION, "HO%d", NUL, NULL },
	{ SET_HARD_POLARITY, "HP%u", NUL, NULL },
	{ SET_HOME_ARMING, "HA%u", NUL, NULL },
	{ SET_HARD_LIMIT, "HL%u", NUL, NULL }, 
	{ SET_HARD_NOTIFY, "HN%u", 'f', NULL }, 
	{ CAPTURE_HOMING_SEQUENCE, "CAHOSEQ", NUL, NULL },
	{ POWER_ON_HOMING_SEQUENCE, "POHOSEQ%u", NUL, NULL },
	{ GO_HOMING_SEQUENCE, "GOHOSEQ", 'h', NULL },	/* ??? */
	{ SET_HOMING_SPEED, "HOSP%u", NUL, NULL },
	{ GET_HOMING_SPEED, "GHOSP", NUL, getIntUnsupported },
	{ SET_HARD_BLOCKING, "HB%u", NUL, NULL }, 
	{ SET_HARD_DIRECTION, "HD%u", NUL, NULL },
	{ GET_HARD_STATUS, "HS", NUL, getIntUnsupported }, 
	{ GET_ACTUAL_HOMING_STATUS, "GAHS", NUL, getHomingStatus },
	{ GET_HOMING_SEQUENCE_CONFIGURATION, "GHSC", NUL, 
	  getIntUnsupported },
	                                        
/* Actual parameters */				
	{ SET_MAXIMUM_SPEED, "SP%u", NUL, NULL },
	{ GET_MAXIMUM_SPEED, "GSP", NUL, getMaxSpeed },
	{ SET_ACCELERATION, "AC%u", NUL, NULL }, 
	{ GET_ACCELERATION, "GAC", NUL, getAcceleration },
	{ SET_PROPORTIONAL_TERM, "POR%d", NUL, NULL },
	{ GET_PROPORTIONAL_TERM, "GPOR", NUL, getPropTerm }, 
	{ SET_INTEGRAL_TERM, "I%d", NUL, NULL },
	{ GET_INTEGRAL_TERM, "GI", NUL, getIntTerm },
	{ SET_PEAK_CURRENT_LIMIT, "LPC%u", NUL, NULL },
	{ GET_PEAK_CURRENT, "GPC", NUL, getIntUnsupported },
	{ SET_CONTINUOUS_CURRENT_LIMIT, "LCC%u", NUL, NULL }, 
	{ GET_CONTINUOUS_CURRENT, "GCC", NUL, getIntUnsupported },
	                                        
/* Additional modes */				
	{ SET_CONTINUOUS_MODE, "CONTMOD", NUL, NULL },
	{ SET_STEPPERMOTOR_MODE, "STEPMOD", NUL, NULL },
	{ SET_STEP_WIDTH, "STW%u", NUL, NULL },
	{ SET_STEP_NUMBER, "STN%u", NUL, NULL },
	{ GET_STEP_WIDTH, "GSTW", NUL, getIntUnsupported },
	{ GET_STEP_NUMBER, "GSTN", NUL, getIntUnsupported },
	{ SET_ANALOG_POS_MODE, "APCMOD", NUL, NULL },
	{ SET_VOLTAGE_MODE, "VOLTMOD", NUL, NULL },
	{ SET_DCR_MODE, "DCRMOD", NUL, NULL }, 
	{ SET_MOTOR_RESISTANCE, "RM%u", NUL, NULL },
	{ SET_SPEED_CONSTANT, "KN%u", NUL, NULL },
	{ GET_MOTOR_RESISTANCE, "GRM", NUL, getIntUnsupported },
	{ GET_SPEED_CONSTANT, "GKN", NUL, getIntUnsupported },
	{ GET_MODE, "GMOD", NUL, getMode }, 
	                                        
/* Error functions and error output */		
	{ SET_DIRECTION_INPUT, "DIRIN", NUL, NULL },
	{ SET_REF_INPUT, "REFIN", NUL, NULL },
	{ SET_ERROR_OUTPUT, "ERROUT", NUL, NULL },
	{ SET_DIGITAL_OUTPUT, "DIGOUT", NUL, NULL },
	{ CLEAR_OUTPUT, "CO", NUL, NULL },
	{ SET_OUTPUT, "SO", NUL, NULL },
	{ SET_DELAYED_CURRENT_ERROR, "DCE%u", NUL, NULL },
	{ GET_DELAYED_CURRENT_ERROR, "GDCE", NUL, getIntUnsupported },
	{ SET_DEVIATION, "DEV%u", NUL, NULL }, 
	{ GET_DEVIATION, "GDEV", NUL, getIntUnsupported },
	                                        
/* Programming sequences */			
	{ PROGRAM_SEQUENCE_BEGIN, "PROGSEQ", NUL, NULL },
	{ PROGRAM_SEQUENCE_END, "END", NUL, NULL },
	{ GET_PROGRAM_SEQUENCE, "GPROGSEQ", NUL, getProgramSequence }, 
	{ ENABLE_PROGRAM, "ENPROG", NUL, NULL }, 
	{ DISABLE_PROGRAM, "DIPROG", NUL, NULL },
	                                        
/* Request Commands */				
	{ GET_ACTUAL_POSITION, "POS", NUL, getPos },
	{ GET_ACTUAL_SPEED, "GN", NUL, getSpeed },
	{ GET_CURRENT_LIMIT, "GCL", NUL, getIntUnsupported },
	{ GET_REAL_CURRENT, "GRC", NUL, getIntUnsupported },
	{ GET_TEMPERATURE, "TEM", NUL, getTemp },
	{ GET_VERSION, "VER", NUL, getVersion },
	{ NOTIFY_ERROR, "NE%u", '?', NULL }, 
	{ GET_STATUS, "GST", NUL, getStatus }, 
	{ GET_FAULT_STATUS, "GFS", NUL, getIntUnsupported },
	{ GET_ACTUAL_STATUS, "GAST", NUL, getActualStatus },
	{ GET_SPECIAL_CONFIG_SET, "GSCS", NUL, getIntUnsupported },
	{ GET_ENHANCED_STATUS, "GES", NUL, getIntUnsupported },
};

#define FH2805_COMMANDS_NUMBER (sizeof(cmdTab)/sizeof(FH2805_CMD_TAB))

/*
 * Handle asynchronous messages 
 */
void
fh2805CheckAsyncAnswer(fh2805dev_t *dev)
{
	if (dev->asyncAnswer != '\0' &&
	    dev->buf[0] == dev->asyncAnswer &&
	    dev->buf[1] == '\r') {
		/* Notify parent */
		fprintf(stderr, "%s: got async %c\n", __func__, dev->buf[0]);
	}
	dev->asyncAnswer = '\0';
}

/*----------------------------------------------------------------------*/
/**
 ** Decoding functions for replies
 ** All this functions first check if the data is an asynchronous message
 ** and then do a specific decoding
 **/

/*
 * A 'catch all' function that read and discard an integer in a reply
 * Used for all requests that we don't handle for now
 */
static int
getIntUnsupported(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	return OK;
}

/*
 * Read the device speed in a reply
 */
static int
getVelocity(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->speed = i;
	return OK;
}

/*
 * Read the homing status -- XXX for now we just discard the data
 */
static int 
getHomingStatus(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	
	return OK;
}

/*
 * read Maximum speed speed 
 */
static int 
getMaxSpeed(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->maxSpeed = i;
	return OK;
}

/*
 * read Acceleration
 */
static int 
getAcceleration(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->acceleration = i;
	return OK;
}

static int 
getPropTerm(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->p = i;
	return OK;
}

static int 
getIntTerm(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->i = i;
	return OK;
}
 
static int
getMode(fh2805dev_t *dev)
{
	fh2805CheckAsyncAnswer(dev);
	return OK;
}

static int
getProgramSequence(fh2805dev_t *dev)
{
	fh2805CheckAsyncAnswer(dev);
	return OK;
}

static int 
getPos(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->pos = i;
	return OK;
}

static int 
getSpeed(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->speed = i;
	return OK;
}

static int 
getTemp(fh2805dev_t *dev)
{
	int i;

	fh2805CheckAsyncAnswer(dev);
	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->temperature = i;
	return OK;
}

static int 
getVersion(fh2805dev_t *dev)
{
	fh2805CheckAsyncAnswer(dev);
	if (strncmp(dev->buf, "Version ", 8) == 0) {
		if (dev->version != NULL) {
			free(dev->version);
		}
		dev->version = strdup(dev->buf);
	} else {
		return ERROR;
	}
	return OK;
}

static int 
getStatus(fh2805dev_t *dev)
{
	int i;

	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	dev->status = i;
	return OK;
}

static int 
getActualStatus(fh2805dev_t *dev)
{
	int i;

	if (sscanf(dev->buf, "%d", &i) != 1) 
		/* out of synch */
		return ERROR;
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** find the structure corrsponding to a given symbolic command
 **/
extern
FH2805_CMD_TAB *
fh2805FindCmd(int cmd)
{
	FH2805_CMD_TAB *c;
	int i;

	if (cmd <= 0 || cmd > FH2805_COMMANDS_NUMBER) {
		return NULL;
	}
	for (c = NULL, i = 0; i < FH2805_COMMANDS_NUMBER; i++) {
		if (cmdTab[i].cmd == cmd) {
			c = &cmdTab[i];
			break;
		}
	}
	return c;
}
